import pytest
import doctest
from insights.parsers.hostname import Hostname as HnF, HostnameShort as HnS, HostnameDefault as HnD
from insights.parsers.facter import Facter
from insights.parsers.systemid import SystemID
from insights.combiners import hostname
from insights.combiners.hostname import Hostname, hostname as hn_func
from insights.tests import context_wrap

HOSTNAME_FULL = "rhel7.example.com"
HOSTNAME_SHORT = "rhel7"
HOSTNAME_DEF = "rhel7"
FACTS_FQDN = """
COMMAND> facter

architecture => x86_64
augeasversion => 1.1.0
facterversion => 1.7.6
filesystems => xfs
fqdn => ewa-satellite.example.com
hostname => ewa-satellite
""".strip()
FACTS_NO_FQDN = """
COMMAND> facter

architecture => x86_64
facterversion => 1.7.6
filesystems => xfs
""".strip()
SYSTEMID_PROFILE_NAME = '''
<?xml version="1.0"?>
<params>
<param>
<value><struct>
<member>
<name>username</name>
<value><string>testuser</string></value>
</member>
<member>
<name>profile_name</name>
<value><string>example_profile</string></value>
</member>
<member>
<name>system_id</name>
<value><string>ID-example</string></value>
</member>
<member>
<name>architecture</name>
<value><string>x86_64</string></value>
</member>
</struct></value>
</param>
</params>
'''.strip()
SYSTEMID_NO_PROFILE_NAME = '''
<?xml version="1.0"?>
<params>
<param>
<member>
<name>username</name>
<value><string>testuser</string></value>
</member>
<member>
<name>architecture</name>
<value><string>x86_64</string></value>
</member>
</param>
</params>
'''.strip()


def test_get_hostname():
    hnf = HnF(context_wrap(HOSTNAME_FULL))
    expected = (HOSTNAME_FULL, HOSTNAME_SHORT, 'example.com')
    result = Hostname(hnf, None, None, None, None)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]

    hns = HnS(context_wrap(HOSTNAME_SHORT))
    expected = (HOSTNAME_SHORT, HOSTNAME_SHORT, '')
    result = Hostname(None, None, hns, None, None)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]

    hnd = HnD(context_wrap(HOSTNAME_DEF))
    expected = (HOSTNAME_DEF, HOSTNAME_DEF, '')
    result = Hostname(None, hnd, None, None, None)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]


def test_get_facter_hostname():
    ft = Facter(context_wrap(FACTS_FQDN))
    expected = ('ewa-satellite.example.com', 'ewa-satellite', 'example.com')
    result = Hostname(None, None, None, ft, None)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]


def test_get_systemid_hostname():
    sid = SystemID(context_wrap(SYSTEMID_PROFILE_NAME))
    expected = ('example_profile', 'example_profile', '')
    result = Hostname(None, None, None, None, sid)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]


def test_get_all_hostname():
    hnf = HnF(context_wrap(HOSTNAME_FULL))
    hns = HnS(context_wrap(HOSTNAME_SHORT))
    hnd = HnD(context_wrap(HOSTNAME_DEF))
    ft = Facter(context_wrap(FACTS_FQDN))
    sid = SystemID(context_wrap(SYSTEMID_PROFILE_NAME))
    expected = (HOSTNAME_FULL, HOSTNAME_SHORT, 'example.com')
    result = Hostname(hnf, hnd, hns, ft, sid)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]


def test_hostname_function():
    # can be removed once the function being removed
    hnf = HnF(context_wrap(HOSTNAME_FULL))
    hns = HnS(context_wrap(HOSTNAME_SHORT))
    hnd = HnD(context_wrap(HOSTNAME_DEF))
    ft = Facter(context_wrap(FACTS_FQDN))
    sid = SystemID(context_wrap(SYSTEMID_PROFILE_NAME))
    expected = (HOSTNAME_FULL, HOSTNAME_SHORT, 'example.com')
    result = hn_func(hnf, hnd, hns, ft, sid)
    assert result.fqdn == expected[0]
    assert result.hostname == expected[1]
    assert result.domain == expected[2]


def test_hostname_raise():
    with pytest.raises(Exception):
        sid = SystemID(context_wrap(SYSTEMID_NO_PROFILE_NAME))
        Hostname(None, None, None, None, sid)


def test_hostname_doc():
    hnf = HnF(context_wrap(HOSTNAME_FULL))
    hns = HnS(context_wrap(HOSTNAME_SHORT))
    hnd = HnD(context_wrap(HOSTNAME_DEF))
    ft = Facter(context_wrap(FACTS_FQDN))
    sid = SystemID(context_wrap(SYSTEMID_PROFILE_NAME))
    env = {
            'hostname': Hostname(hnf, hnd, hns, ft, sid),
            'hn': hn_func(hnf, hnd, hns, ft, sid)
          }
    failed, total = doctest.testmod(hostname, globs=env)
    assert failed == 0



"""
Octavia - file ``octavia.conf``
===============================

Provides a parser arser for file
``/var/lib/config-data/puppet-generated/octavia/etc/octavia/octavia.conf``.  Filters
have been added to this parser to ensure that the necessary data will be collected.

Sample input data::

    [DEFAULT]
    # Print debugging output (set logging level to DEBUG instead of default WARNING level).
    debug = False

    # Plugin options are hot_plug_plugin (Hot-pluggable controller plugin)
    # octavia_plugins = hot_plug_plugin

    # Hostname to be used by the host machine for services running on it.
    # The default value is the hostname of the host machine.
    # host =

    # AMQP Transport URL
    # For Single Host, specify one full transport URL:
    #   transport_url = rabbit://<user>:<pass>@127.0.0.1:5672/<vhost>
    # For HA, specify queue nodes in cluster, comma delimited:
    #   transport_url = rabbit://<user>:<pass>@server01,<user>:<pass>@server02/<vhost>
    # transport_url =

    # How long in seconds to wait for octavia worker to exit before killing them.
    # graceful_shutdown_timeout = 60

    [api_settings]
    bind_host = 127.0.0.1
    bind_port = 9876

    # How should authentication be handled (keystone, noauth)
    # auth_strategy = keystone

    # allow_pagination = True
    # allow_sorting = True
    # pagination_max_limit = 1000
    # Base URI for the API for use in pagination links.
    # This will be autodetected from the request if not overridden here.
    # Example:
    #   api_base_uri = http://localhost:9876
    # api_base_uri =

    # Enable/disable ability for users to create TLS Terminated listeners
    # allow_tls_terminated_listeners = True

    # Enable/disable ability for users to create PING type Health Monitors
    # allow_ping_health_monitors = True

    # Dictionary of enabled provider driver names and descriptions
    # A comma separated list of dictionaries of the enabled provider driver names
    # and descriptions.
    # enabled_provider_drivers = amphora:The Octavia Amphora driver.,octavia: \
    #                            Deprecated alias of the Octavia Amphora driver.

    # Default provider driver
    default_provider_driver = amphora

    # The minimum health monitor delay interval for UDP-CONNECT Health Monitor type
    udp_connect_min_interval_health_monitor = 3

Examples:
    >>> type(octavia_conf)
    <class 'insights.parsers.octavia.OctaviaConf'>
    >>> octavia_conf.defaults()['debug'] == 'False'
    True
    >>> octavia_conf.get('api_settings', 'bind_port') == '9876'
    True
    >>> octavia_conf.has_option('api_settings', 'missing_key')
    False
"""
from insights import IniConfigFile, parser
from insights.core.filters import add_filter
from insights.specs import Specs

VALID_KEYS = [
    '[',
    'active_connection_max_retries',
    'active_connection_rety_interval',
    'admin_log_targets',
    'administrative_log_facility',
    'agent_request_read_timeout',
    'agent_server_ca',
    'agent_server_cert',
    'agent_server_network_dir',
    'agent_server_network_file',
    'agent_tls_protocol',
    'allow_pagination',
    'allow_ping_health_monitors',
    'allow_sorting',
    'allow_tls_terminated_listeners',
    'allow_vip_network_id',
    'allow_vip_port_id',
    'allow_vip_subnet_id',
    'amp_active_retries',
    'amp_active_wait_sec',
    'amp_boot_network_list',
    'amp_flavor_id',
    'amp_image_id',
    'amp_image_owner_id',
    'amp_image_tag',
    'amp_secgroup_list',
    'amp_ssh_access_allowed',
    'amphora_driver',
    'amphora_expiry_age',
    'amphora_udp_driver',
    'anti_affinity_policy',
    'api_base_uri',
    'audit_map_file',
    'auth_strategy',
    'auth_type',
    'availability_zone',
    'base_cert_dir',
    'base_path',
    'bind_host',
    'bind_ip',
    'bind_port',
    'build_active_retries',
    'build_rate_limit',
    'build_retry_interval',
    'ca_certificates_file',
    'cafile',
    'cert_generator',
    'cert_manager',
    'cert_validity_time',
    'cleanup_interval',
    'client_ca',
    'client_cert',
    'compute_driver',
    'connection_logging',
    'connection_max_retries',
    'connection_retry_interval',
    'controller_ip_port_list',
    'debug',
    'default_health_monitor_quota',
    'default_listener_quota',
    'default_load_balancer_quota',
    'default_member_quota',
    'default_pool_quota',
    'default_provider_driver',
    'disable_local_log_storage',
    'disable_revert',
    'distributor_driver',
    'driver',
    'enable_anti_affinity',
    'enable_proxy_headers_parsing',
    'enabled',
    'enabled_provider_agents',
    'enabled_provider_drivers',
    'endpoint_type',
    'engine',
    'failover_threads',
    'forward_all_logs',
    'get_socket_path',
    'graceful_shutdown_timeout',
    'haproxy_cmd',
    'haproxy_stick_size',
    'haproxy_template',
    'health_check_interval',
    'health_update_driver',
    'health_update_threads',
    'heartbeat_interval',
    'heartbeat_timeout',
    'ignore_req_list',
    'insecure',
    'lb_network_interface',
    'load_balancer_expiry_age',
    'loadbalancer_topology',
    'log_dir',
    'log_file',
    'log_protocol',
    'log_queue_size',
    'log_retry_count',
    'log_retry_interval',
    'logging_template_override',
    'max_process_warning_percent',
    'max_retries',
    'max_workers',
    'memcached_servers',
    'network_driver',
    'octavia_plugins',
    'pagination_max_limit',
    'policy_file',
    'port_detach_timeout',
    'provider_agent_shutdown_timeout',
    'random_amphora_name_length',
    'region_name',
    'respawn_count',
    'respawn_interval',
    'rest_request_conn_timeout',
    'rest_request_read_timeout',
    'retry_interval',
    'rpc_thread_pool_size',
    'server_ca',
    'service_name',
    'signing_digest',
    'sock_rlimit',
    'spare_amphora_pool_size',
    'spare_check_interval',
    'stats_max_processes',
    'stats_request_timeout',
    'stats_socket_path',
    'stats_update_driver',
    'stats_update_threads',
    'status_max_processes',
    'status_request_timeout',
    'status_socket_path',
    'status_update_threads',
    'storage_path',
    'tenant_log_targets',
    'topic',
    'topics',
    'udp_connect_min_interval_health_monitor',
    'use_oslo_messaging',
    'use_upstart',
    'user_data_config_drive',
    'user_log_facility',
    'user_log_format',
    'volume_create_max_retries',
    'volume_create_retry_interval',
    'volume_create_timeout',
    'volume_driver',
    'volume_size',
    'volume_type',
    'vrrp_advert_int',
    'vrrp_check_interval',
    'vrrp_fail_count',
    'vrrp_garp_refresh_count',
    'vrrp_garp_refresh_interval',
    'vrrp_success_count',
    'workers'
]

add_filter(Specs.octavia_conf, VALID_KEYS)


@parser(Specs.octavia_conf)
class OctaviaConf(IniConfigFile):
    """
    Parser for file ``/var/lib/config-data/puppet-generated/octavia/etc/octavia/octavia.conf``
    """
    pass



import json
import logging
import os

from kubernetes import config
from kubernetes.client import ApiClient, Configuration
from openshift.dynamic import DynamicClient

from insights.core.context import ExecutionContext
from insights.core.plugins import datasource
from insights.core.serde import deserializer, serializer
from insights.core.spec_factory import ContentProvider, SerializedRawOutputProvider
from insights.util import fs

log = logging.getLogger(__name__)


class GVK(object):
    def __init__(self, kind, api_version="v1", kwargs=None):
        self.kind = kind
        self.api_version = api_version
        self.kwargs = kwargs or {}


class OpenshiftOutputProvider(ContentProvider):
    def __init__(self, client, kind=None, api_version=None, **kwargs):
        super(OpenshiftOutputProvider, self).__init__()
        self.kind = kind
        self.api_version = api_version
        name = "%s/%s" % (self.api_version, self.kind)
        self.gvk = name.split("/")
        self.relative_path = name
        self.root = "/"
        self.kwargs = kwargs
        self.k8s = client.k8s

    def load(self):
        return self.k8s.resources.get(kind=self.kind, api_version=self.api_version).get(serialize=False, **self.kwargs).data

    def write(self, dst):
        fs.ensure_path(os.path.dirname(dst))
        with open(dst, "wb") as f:
            f.write(self.content)

        # we're done with it if we're writing it down.
        # reset _content so we don't build up memory
        self.loaded = False
        self._content = None


@serializer(OpenshiftOutputProvider)
def serialize_openshift_output(obj, root):
    rel = os.path.join("k8s", *obj.gvk)
    dst = os.path.join(root, rel)
    fs.ensure_path(os.path.dirname(dst))
    obj.write(dst)
    return {"relative_path": rel}


@deserializer(OpenshiftOutputProvider)
def deserialize_openshift_output(_type, data, root):
    rel = data["relative_path"]
    res = SerializedRawOutputProvider(rel, root)
    return res


class OpenshiftContext(ExecutionContext):
    pass


class OpenshiftClient(object):
    def __init__(self, ctx=None, cfg=None):
        cfg = cfg or os.environ.get("KUBECONFIG")
        if cfg:
            k8s_client = config.new_client_from_config(cfg)
        else:
            config.load_incluster_config()  # makes a singleton config behind the scenes
            k8cfg = Configuration()  # gets a copy from what was populated in the line above
            # NOTE this is required due to https://github.com/openshift/origin/issues/22125
            k8cfg.verify_ssl = False
            k8s_client = ApiClient(configuration=k8cfg)  # this should use the singleton produced above
        self.k8s = DynamicClient(k8s_client)  # stole this from config.new_client_from_config


client = OpenshiftClient()


class resource(object):
    client_kwargs = None
    timeout = None

    def __init__(self, kind, api_version="v1", **kwargs):
        # encode group into the api_version string if necessary
        self.static_kwargs = kwargs
        self.kind = kind
        self.api_version = api_version
        self.__name__ = self.__class__.__name__
        datasource(OpenshiftContext)(self)

    def __call__(self, broker):
        # allow manifest to override what's in the resource definition
        ctx = broker[OpenshiftContext]
        timeout = self.timeout or ctx.timeout
        kwargs = dict(self.client_kwargs if self.client_kwargs is not None else self.static_kwargs)
        if timeout:
            kwargs["timeout_seconds"] = timeout
        return OpenshiftOutputProvider(client, kind=self.kind, api_version=self.api_version, **kwargs)


class foreach_resource(object):
    client_kwargs = None
    timeout = None

    def __init__(self, dep, func):
        self.dep = dep
        self.func = func
        self.__name__ = self.__class__.__name__
        datasource(OpenshiftContext, dep)(self)

    def __call__(self, broker):
        # allow manifest to override what's in the resource definition
        ctx = broker[OpenshiftContext]
        timeout = self.timeout or ctx.timeout
        crds = broker[self.dep]
        _lst = []
        for crd in json.loads(crds.content)["items"]:
            gvk = self.func(crd)
            kwargs = dict(self.client_kwargs if self.client_kwargs is not None else gvk.kwargs)
            if timeout:
                kwargs["timeout_seconds"] = timeout
            _lst.append(OpenshiftOutputProvider(client, kind=gvk.kind, api_version=gvk.api_version, **kwargs))
        return _lst



import pytest
from collections import OrderedDict
from insights.parsers import split_kv_pairs, unsplit_lines, parse_fixed_table
from insights.parsers import calc_offset, optlist_to_dict, keyword_search
from insights.parsers import parse_delimited_table, ParseException, SkipException

SPLIT_TEST_1 = """
# Comment line

  keyword1 = value1   # Inline comments

  # Comment indented
  keyword3     # Key with no separator
  keyword2 = value2a=True, value2b=100M

""".strip()

SPLIT_TEST_1_OD = OrderedDict([
    ('keyword1', 'value1'),
    ('keyword3', ''),
    ('keyword2', 'value2a=True, value2b=100M')
])

SPLIT_TEST_2 = """
@ Comment line

  keyword1: value1   @ Inline comments
  keyword2 : value2a=True, value2b=100M

  @ Comment indented
  keyword3     @ Key with no separator
""".strip()

OFFSET_CONTENT_1 = """
  data 1 line
data 2 line
""".strip()

OFFSET_CONTENT_2 = """
#
Warning line
Error line
    data 1 line
    data 2 line
Trailing line

Blank line above
Another trailing line
    Yet another trailing line
    Yet yet another trailing line
""".strip()


def test_split_kv_pairs():
    kv_pairs = split_kv_pairs(SPLIT_TEST_1.splitlines())
    assert len(kv_pairs) == 2
    assert kv_pairs == {
        'keyword1': 'value1',
        'keyword2': 'value2a=True, value2b=100M'
    }

    kv_pairs = split_kv_pairs(SPLIT_TEST_1.splitlines(), filter_string='value2')
    assert len(kv_pairs) == 1
    assert kv_pairs == {
        'keyword2': 'value2a=True, value2b=100M'
    }

    kv_pairs = split_kv_pairs(SPLIT_TEST_1.splitlines(), use_partition=True)
    assert len(kv_pairs) == 3
    assert kv_pairs == {
        'keyword1': 'value1',
        'keyword2': 'value2a=True, value2b=100M',
        'keyword3': ''
    }

    kv_pairs = split_kv_pairs(SPLIT_TEST_1.splitlines(), use_partition=True, ordered=True)
    assert len(kv_pairs) == 3
    assert kv_pairs == SPLIT_TEST_1_OD

    kv_pairs = split_kv_pairs(SPLIT_TEST_2.splitlines(), comment_char='@', split_on=':')
    assert len(kv_pairs) == 2
    assert kv_pairs == {
        'keyword1': 'value1',
        'keyword2': 'value2a=True, value2b=100M'
    }

    kv_pairs = split_kv_pairs(SPLIT_TEST_2.splitlines(), comment_char='@', split_on=':', filter_string='value2')
    assert len(kv_pairs) == 1
    assert kv_pairs == {
        'keyword2': 'value2a=True, value2b=100M'
    }

    kv_pairs = split_kv_pairs(SPLIT_TEST_2.splitlines(), comment_char='@', split_on=':', use_partition=True)
    assert len(kv_pairs) == 3
    assert kv_pairs == {
        'keyword1': 'value1',
        'keyword2': 'value2a=True, value2b=100M',
        'keyword3': ''
    }


SPLIT_LINES = """
Line one
Line two part 1 \\
         line two part 2\\
         line two part 3
Line three
""".strip()
SPLIT_LINES_2 = """
Line one
Line two part 1 ^
         line two part 2^
         line two part 3
Line three^
""".strip()
SPLIT_LINES_3 = """
web.default_taskmaster_tasks = RHN::Task::SessionCleanup, RHN::Task::ErrataQueue,
    RHN::Task::ErrataEngine,
    RHN::Task::DailySummary, RHN::Task::SummaryPopulation,
    RHN::Task::RHNProc,
    RHN::Task::PackageCleanup

db_host ="""


def test_unsplit_lines():
    lines = list(unsplit_lines(SPLIT_LINES.splitlines()))
    assert len(lines) == 3
    assert lines[0] == 'Line one'
    assert lines[1] == 'Line two part 1          line two part 2         line two part 3'
    assert lines[2] == 'Line three'

    lines = list(unsplit_lines(SPLIT_LINES_2.splitlines(), cont_char='^'))
    assert len(lines) == 3
    assert lines[0] == 'Line one'
    assert lines[1] == 'Line two part 1          line two part 2         line two part 3'
    assert lines[2] == 'Line three'  # test continuation on last line

    # Test keeping continuation character on line
    lines = list(unsplit_lines(
        SPLIT_LINES_3.splitlines(), cont_char=',', keep_cont_char=True
    ))
    assert len(lines) == 4
    assert lines[0] == ''
    assert lines[1] == 'web.default_taskmaster_tasks = RHN::Task::SessionCleanup, RHN::Task::ErrataQueue,    RHN::Task::ErrataEngine,    RHN::Task::DailySummary, RHN::Task::SummaryPopulation,    RHN::Task::RHNProc,    RHN::Task::PackageCleanup'
    assert lines[2] == ''
    assert lines[3] == 'db_host ='


def test_calc_offset():
    assert calc_offset(OFFSET_CONTENT_1.splitlines(), target=[]) == 0
    assert calc_offset(OFFSET_CONTENT_1.splitlines(), target=[None]) == 0
    assert calc_offset(OFFSET_CONTENT_1.splitlines(), target=['data ']) == 0
    with pytest.raises(ValueError):
        calc_offset(OFFSET_CONTENT_1.splitlines(), target=['xdata '])
    with pytest.raises(ValueError):
        calc_offset(OFFSET_CONTENT_1.splitlines(),
                    target=['data '],
                    invert_search=True)
    assert calc_offset(OFFSET_CONTENT_1.splitlines(),
                       target=['Trailing', 'Blank', 'Another '],
                       invert_search=True) == 0
    assert calc_offset(OFFSET_CONTENT_2.splitlines(), target=[]) == 0
    assert calc_offset(OFFSET_CONTENT_2.splitlines(), target=['data ']) == 3
    assert calc_offset(reversed(OFFSET_CONTENT_2.splitlines()),
                       target=['Trailing', 'Blank', 'Another ', 'Yet'],
                       invert_search=True) == 6


FIXED_CONTENT_1 = """
Column1    Column2    Column3
data1      data 2     data   3
     data4 data5      data6
data     7            data   9
""".strip()

FIXED_CONTENT_1A = """
    WARNING
    Column1    Column2    Column3
    data1      data 2     data   3
         data4 data5      data6
    data     7            data   9
""".strip()

FIXED_CONTENT_1B = """
Column1    Column2    Column3
data1      data 2
data4      data5      data6
  data   7            data   9
""".strip()

FIXED_CONTENT_2 = """
WARNING WARNING WARNING
 Some message
Another message
Column1    Column2    Column3
data1      data 2     data   3
     data4 data5      data6
data     7            data   9
""".strip()

FIXED_CONTENT_3 = """
WARNING WARNING WARNING
 Some message
Another message
Column1    Column2    Column3
data1      data 2     data   3
     data4 data5      data6
data     7            data   9
Trailing non-data line
 Another trailing non-data line
""".strip()

FIXED_CONTENT_4 = """
WARNING WARNING WARNING
 Some message
Another message
Column1    Column 2    Column 3
data1      data 2      data   3
     data4 data5       data6
data     7             data   9
data10
Trailing non-data line
 Another trailing non-data line
""".strip()

FIXED_CONTENT_5 = """
Column1    Column 2    Column 3

data1      data 2      data   3

data       7             data   9

data10
""".strip()


FIXED_CONTENT_DUP_HEADER_PREFIXES = """
NAMESPACE    NAME    LABELS
default      foo     app=superawesome
""".strip()


def test_parse_fixed_table():
    data = parse_fixed_table(FIXED_CONTENT_1.splitlines())
    assert len(data) == 3
    assert data[0] == {'Column1': 'data1', 'Column2': 'data 2', 'Column3': 'data   3'}
    assert data[1] == {'Column1': 'data4', 'Column2': 'data5', 'Column3': 'data6'}
    assert data[2] == {'Column1': 'data     7', 'Column2': '', 'Column3': 'data   9'}

    data = parse_fixed_table(FIXED_CONTENT_1A.splitlines(), heading_ignore=['Column1 '])
    assert len(data) == 3
    assert data[0] == {'Column1': 'data1', 'Column2': 'data 2', 'Column3': 'data   3'}
    assert data[1] == {'Column1': 'data4', 'Column2': 'data5', 'Column3': 'data6'}
    assert data[2] == {'Column1': 'data     7', 'Column2': '', 'Column3': 'data   9'}

    data = parse_fixed_table(FIXED_CONTENT_1B.splitlines())
    assert len(data) == 3
    assert data[0] == {'Column1': 'data1', 'Column2': 'data 2', 'Column3': ''}
    assert data[1] == {'Column1': 'data4', 'Column2': 'data5', 'Column3': 'data6'}
    assert data[2] == {'Column1': 'data   7', 'Column2': '', 'Column3': 'data   9'}

    data = parse_fixed_table(FIXED_CONTENT_2.splitlines(), heading_ignore=['Column1 '])
    assert len(data) == 3
    assert data[0] == {'Column1': 'data1', 'Column2': 'data 2', 'Column3': 'data   3'}
    assert data[1] == {'Column1': 'data4', 'Column2': 'data5', 'Column3': 'data6'}
    assert data[2] == {'Column1': 'data     7', 'Column2': '', 'Column3': 'data   9'}

    data = parse_fixed_table(FIXED_CONTENT_3.splitlines(),
                             heading_ignore=['Column1 '],
                             trailing_ignore=['Trailing', 'Another'])
    assert len(data) == 3
    assert data[0] == {'Column1': 'data1', 'Column2': 'data 2', 'Column3': 'data   3'}
    assert data[1] == {'Column1': 'data4', 'Column2': 'data5', 'Column3': 'data6'}
    assert data[2] == {'Column1': 'data     7', 'Column2': '', 'Column3': 'data   9'}

    data = parse_fixed_table(FIXED_CONTENT_4.splitlines(),
                             heading_ignore=['Column1 '],
                             header_substitute=[('Column 2', 'Column_2'), ('Column 3', 'Column_3')],
                             trailing_ignore=['Trailing', 'Another'])
    assert len(data) == 4
    assert data[0] == {'Column1': 'data1', 'Column_2': 'data 2', 'Column_3': 'data   3'}
    assert data[1] == {'Column1': 'data4', 'Column_2': 'data5', 'Column_3': 'data6'}
    assert data[2] == {'Column1': 'data     7', 'Column_2': '', 'Column_3': 'data   9'}
    assert data[3] == {'Column1': 'data10', 'Column_2': '', 'Column_3': ''}

    # Test that if we search for trailing data that is always found, then we
    # should get the whole thing parsed as a table from the header line
    data = parse_fixed_table(
        ['foo' + line for line in FIXED_CONTENT_4.splitlines()],
        heading_ignore=['fooColumn1 '],
        header_substitute=[('fooColumn1', 'Column1'), ('Column 2', 'Column_2'), ('Column 3', 'Column_3')],
        trailing_ignore=['foo']
    )
    assert len(data) == 6
    assert data[4] == {'Column1': 'fooTrailing', 'Column_2': 'non-data li', 'Column_3': 'ne'}
    assert data[5] == {'Column1': 'foo Another', 'Column_2': 'trailing no', 'Column_3': 'n-data line'}

    data = parse_fixed_table(FIXED_CONTENT_DUP_HEADER_PREFIXES.splitlines())
    assert data[0] == {'NAMESPACE': 'default', 'NAME': 'foo', 'LABELS': 'app=superawesome'}
    data = parse_fixed_table(FIXED_CONTENT_5.splitlines())
    assert len(data) == 3


def test_parse_fixed_table_empty_exception():
    with pytest.raises(ParseException) as pe:
        parse_fixed_table(FIXED_CONTENT_1B.splitlines(), empty_exception=True)
    assert "Incorrect line:" in str(pe.value)


def test_optlist_standard():
    d = optlist_to_dict('key1,key2=val2,key1=val1,key3')
    assert sorted(d.keys()) == sorted(['key1', 'key2', 'key3'])
    assert d['key1'] == 'val1'
    assert d['key2'] == 'val2'
    assert d['key3'] is True


def test_optlist_no_vals():
    d = optlist_to_dict('key1,key2=val2,key1=val1,key3', kv_sep=None)
    assert sorted(d.keys()) == sorted(['key1', 'key1=val1', 'key2=val2', 'key3'])
    assert d['key1'] is True
    assert d['key1=val1'] is True
    assert d['key2=val2'] is True
    assert d['key3'] is True


def test_optlist_strip_quotes():
    d = optlist_to_dict(
        '''key1="foo",key2='bar',key3="mismatched quotes',key4="inner'quotes"''',
        strip_quotes=True
    )
    assert sorted(d.keys()) == sorted(['key1', 'key2', 'key3', 'key4'])
    assert d['key1'] == 'foo'
    assert d['key2'] == 'bar'
    assert d['key3'] == '"mismatched quotes\''
    assert d['key4'] == "inner'quotes"


def test_optlist_with_spaces():
    d = optlist_to_dict(
        '''key1=foo,  key2=bar'''
    )
    assert 'key1' in d
    assert 'key2' in d


PS_AUX_TEST = """
USER       PID %CPU %MEM    VSZ   RSS TTY      STAT START   TIME COMMAND
root         1  0.0  0.0  19356  1544 ?        Ss   May31   0:01 /sbin/init
root      1821  0.0  0.0      0     0 ?        S    May31   0:25 [kondemand/0]
root      1864  0.0  0.0  18244   668 ?        Ss   May31   0:05 irqbalance --pid=/var/run/irqbalance.pid
user1    20160  0.0  0.0 108472  1896 pts/3    Ss   10:09   0:00 bash
root     20357  0.0  0.0   9120   760 ?        Ss   10:09   0:00 /sbin/dhclient -1 -q -lf /var/lib/dhclient/dhclient-extbr0.leases -pf /var/run/dhclient-extbr0.pid extbr0
qemu     22673  0.8 10.2 1618556 805636 ?      Sl   11:38   1:07 /usr/libexec/qemu-kvm -name rhel7 -S -M rhel6.5.0 -enable-kvm -m 1024 -smp 2,sockets=2,cores=1,threads=1 -uuid 13798ffc-bc1e-d437-4f3f-2e0fa6c923ad
"""

MISSING_DATA_TEST = """
  WARNING: Locking disabled. Be careful! This could corrupt your metadata.
LVM2_PV_FMT|LVM2_PV_UUID|LVM2_DEV_SIZE|LVM2_PV_NAME|LVM2_PV_MAJOR|LVM2_PV_MINOR|LVM2_PV_MDA_FREE|LVM2_PV_MDA_SIZE|LVM2_PV_EXT_VSN|LVM2_PE_START|LVM2_PV_SIZE|LVM2_PV_FREE|LVM2_PV_USED|LVM2_PV_ATTR|LVM2_PV_ALLOCATABLE|LVM2_PV_EXPORTED|LVM2_PV_MISSING|LVM2_PV_PE_COUNT|LVM2_PV_PE_ALLOC_COUNT|LVM2_PV_TAGS|LVM2_PV_MDA_COUNT|LVM2_PV_MDA_USED_COUNT|LVM2_PV_BA_START|LVM2_PV_BA_SIZE|LVM2_PV_IN_USE|LVM2_PV_DUPLICATE|LVM2_VG_NAME
  WARNING: Locking disabled. Be careful! This could corrupt your metadata.
"""

SUBSTITUTE_HEADERS_TEST = """
address,port,state,read-only
0.0.0.0,3000,LISTEN,N
10.76.19.184,37500,ESTAB,Y
""".strip()

POSTGRESQL_LOG = """
 schema |             table              |   rows
 public | rhnsnapshotpackage             | 47428950
 public | rhnpackagefile                 | 32174333
 public | rhnpackagecapability           | 12934215
 public | rhnpackagechangelogrec         | 11269933
 public | rhnchecksum                    | 10129746
 public | rhnactionconfigrevision        |  2894957
 public | rhnpackageprovides             |  2712442
 public | rhnpackagerequires             |  2532861
 public | rhn_command_target             |  1009152
 public | rhnconfigfilename              |        0
 public | rhnxccdfidentsystem            |        0
 public | rhndistchannelmap              |        0
 public | rhnactionvirtshutdown          |        0
 public | rhnpublicchannelfamily         |        0
 (402 rows)
""".strip()  # Normally has a --- separator line, which is ignored using get_active_lines

TABLE1 = """
THIS   IS   A   HEADER
 this   is   some   content_with_blank_prefix
This   is   more   content
""".strip()

TABLE2 = [
    "SID   Nr   Instance    SAPLOCALHOST                        Version                 DIR_EXECUTABLE",
    "HA2|  16|       D16|         lu0417|749, patch 10, changelist 1698137|          /usr/sap/HA2/D16/exe",
    "HA2|  22|       D22|         lu0417|749, patch 10, changelist 1698137|          /usr/sap/HA2/D22/exe"
]

TABLE3 = """
THIS | IS | A | HEADER
this ^ is ^ some ^ content
This ^ is ^ more ^ content
""".strip()


def test_parse_delimited_table():
    # No content?  No table.
    assert parse_delimited_table([]) == []

    # Test maximum splits and header 'ignore', which should actually be
    # called 'header_startswith'
    tbl = parse_delimited_table(
        PS_AUX_TEST.splitlines(), max_splits=10, heading_ignore=['USER']
    )
    assert tbl
    assert isinstance(tbl, list)
    assert len(tbl) == 6
    assert isinstance(tbl[0], dict)
    assert tbl[0] == {
        '%MEM': '0.0', 'TTY': '?', 'VSZ': '19356', 'PID': '1', '%CPU': '0.0',
        'START': 'May31', 'COMMAND': '/sbin/init', 'USER': 'root',
        'STAT': 'Ss', 'TIME': '0:01', 'RSS': '1544'
    }
    assert tbl[5]['COMMAND'] == \
        '/usr/libexec/qemu-kvm -name rhel7 -S -M rhel6.5.0 -enable-kvm -m 1024 -smp 2,sockets=2,cores=1,threads=1 -uuid 13798ffc-bc1e-d437-4f3f-2e0fa6c923ad'

    # Test trailing ignore not found
    tbl = parse_delimited_table(
        MISSING_DATA_TEST.splitlines(), delim='|',
        heading_ignore=['LVM2_PV_FMT'],
        trailing_ignore=['WARNING', 'ERROR', 'Cannot get lock']
    )
    assert isinstance(tbl, list)
    assert len(tbl) == 0

    # Header substitution
    tbl = parse_delimited_table(
        SUBSTITUTE_HEADERS_TEST.splitlines(), delim=',', strip=False,
        header_substitute=[('read-only', 'read_only')]
    )
    assert tbl
    assert isinstance(tbl, list)
    assert len(tbl) == 2
    assert isinstance(tbl[1], dict)
    assert tbl[1] == {
        'address': '10.76.19.184', 'port': '37500', 'state': 'ESTAB', 'read_only': 'Y'
    }

    # Test change of delimiter and trailing_ignore
    tbl = parse_delimited_table(POSTGRESQL_LOG.splitlines(), delim='|', trailing_ignore=['('])
    assert isinstance(tbl, list)
    assert len(tbl) == 14
    assert isinstance(tbl[0], dict)
    assert tbl[0] == {
        'schema': 'public', 'table': 'rhnsnapshotpackage', 'rows': '47428950'
    }

    # Test using different header delimiter
    result = parse_delimited_table(TABLE3.splitlines(), delim="^", header_delim="|")
    assert isinstance(result, list)
    assert len(result) == 2
    assert isinstance(result[0], dict)
    expected = [{"THIS": "this", "IS": "is", "A": "some", "HEADER": "content"},
                {"THIS": "This", "IS": "is", "A": "more", "HEADER": "content"}]
    assert expected == result

    # Test explicit None as header delimiter, different from content delimiter
    result = parse_delimited_table(TABLE2, delim='|', header_delim=None)
    assert isinstance(result, list)
    assert len(result) == 2
    assert isinstance(result[0], dict)
    expected = [{"SID": "HA2", "Nr": "16", "Instance": "D16", "SAPLOCALHOST": "lu0417",
                 "Version": "749, patch 10, changelist 1698137",
                 "DIR_EXECUTABLE": "/usr/sap/HA2/D16/exe"},
                {"SID": "HA2", "Nr": "22", "Instance": "D22", "SAPLOCALHOST": "lu0417",
                 "Version": "749, patch 10, changelist 1698137",
                 "DIR_EXECUTABLE": "/usr/sap/HA2/D22/exe"}]
    assert expected == result

    # Test raw_line_key
    TABLE1_SP = TABLE1.splitlines()
    result = parse_delimited_table(TABLE1_SP, raw_line_key='raw_line')
    assert isinstance(result, list)
    assert len(result) == 2
    assert isinstance(result[0], dict)
    # Get the RAW line
    assert result[0]['raw_line'] == TABLE1_SP[1]


DATA_LIST = [
    {'name': 'test 1', 'role': 'server', 'memory_gb': 16, 'ssd': True},
    {'name': 'test 2', 'role': 'server', 'memory_gb': 256, 'ssd': False},
    {'name': 'test 3', 'role': 'server', 'memory_gb': 16, 'ssd': False},
    {'name': 'test 4', 'role': 'embedded', 'memory_gb': 1, 'ssd': False},
    {'name': 'test 5', 'role': 'workstation', 'memory_gb': 16, 'ssd': True},
]

CERT_LIST = [
    {
        'status': 'MONITORING',
        'stuck': 'no',
        'key pair storage': "type=NSSDB,location='/etc/dirsrv/slapd-LDAP-EXAMPLE-COM',nickname='Server-Cert',token='NSS Certificate DB',pinfile='/etc/dirsrv/slapd-LDAP-EXAMPLE-COM/pwdfile.txt'",
        'certificate': {
            'type': 'NSSDB',
            'location': '/etc/dirsrv/slapd-LDAP-EXAMPLE-COM',
            'nickname': 'Server-Cert',
            'token': 'NSS Certificate DB',
        },
        'CA': 'IPA',
        'issuer': 'CN=Certificate Authority,O=LDAP.EXAMPLE.COM',
        'subject': 'CN=master.LDAP.EXAMPLE.COM,O=LDAP.EXAMPLE.COM',
        'expires': '2017-06-28 12:52:12 UTC',
        'eku': 'id-kp-serverAuth,id-kp-clientAuth',
        'pre-save command': '',
        'post-save command': '/usr/lib64/ipa/certmonger/restart_dirsrv LDAP-EXAMPLE-COM',
        'track': 'yes',
        'auto-renew': 'yes',
    }, {
        'status': 'MONITORING',
        'stuck': 'no',
        'key pair storage': "type=NSSDB,location='/etc/dirsrv/slapd-PKI-IPA',nickname='Server-Cert',token='NSS Certificate DB',pinfile='/etc/dirsrv/slapd-PKI-IPA/pwdfile.txt'",
        'certificate': {
            'type': 'NSSDB',
            'location': '/etc/dirsrv/slapd-PKI-IPA',
            'nickname': 'Server-Cert',
            'token': 'NSS Certificate DB',
        },
        'CA': 'IPA',
        'issuer': 'CN=Certificate Authority,O=EXAMPLE.COM',
        'subject': 'CN=ldap.EXAMPLE.COM,O=EXAMPLE.COM',
        'expires': '2017-06-28 12:52:13 UTC',
        'eku': 'id-kp-serverAuth,id-kp-clientAuth',
        'pre-save command': '',
        'post-save command': '/usr/lib64/ipa/certmonger/restart_dirsrv PKI-IPA',
        'track': 'yes',
        'auto-renew': 'yes',
        'dash- space': 'tested',
    }
]


def test_keyword_search():
    # No keywords, no result
    assert len(keyword_search(DATA_LIST)) == 0
    # Search on absent keywords produces empty list
    assert keyword_search(DATA_LIST, cpu_count=4) == []
    # Search on present but non-matching keyword produces empty list
    assert keyword_search(DATA_LIST, memory_gb=8) == []
    # Single result - search on string
    results = keyword_search(DATA_LIST, role='embedded')
    assert len(results) == 1
    assert results[0] == DATA_LIST[3]
    # Multiple results, name has underscore - search on integer
    results = keyword_search(DATA_LIST, memory_gb=16)
    assert len(results) == 3
    assert results == [DATA_LIST[i] for i in (0, 2, 4)]
    # Search on boolean
    results = keyword_search(DATA_LIST, ssd=False)
    assert len(results) == 3
    assert results == [DATA_LIST[i] for i in (1, 2, 3)]
    # No data, no results.
    assert len(keyword_search([], role='server')) == 0

    # Search with contains
    results = keyword_search(DATA_LIST, role__contains='e')
    assert len(results) == 4
    assert results == [DATA_LIST[i] for i in (0, 1, 2, 3)]

    # Search with startswith
    results = keyword_search(DATA_LIST, role__startswith='e')
    assert len(results) == 1
    assert results[0] == DATA_LIST[3]

    # Search for multiple keys, with spaces and dashes, and search operators
    results = keyword_search(
        CERT_LIST,
        pre_save_command='',
        key_pair_storage__startswith="type=NSSDB,location='/etc/dirsrv/slapd-PKI-IPA'"
    )
    assert len(results) == 1
    assert results[0] == CERT_LIST[1]

    # Make sure contains can also apply to keys with dashes and spaces
    results = keyword_search(
        CERT_LIST,
        post_save_command__contains='PKI-IPA',
    )
    assert len(results) == 1
    assert results[0] == CERT_LIST[1]

    # Lower case value matching
    results = keyword_search(
        CERT_LIST,
        status__lower_value='Monitoring',
    )
    assert len(results) == 2
    assert results == CERT_LIST

    # Check that searches for keys with two underscores that aren't matcher
    # suffixes still work
    results = keyword_search(
        CERT_LIST,
        dash__space='tested',
    )
    assert len(results) == 1
    assert results[0] == CERT_LIST[1]

    # Check that we can use contains to check the contents of a dictionary
    # in a value
    results = keyword_search(
        CERT_LIST,
        certificate__contains='type'
    )
    assert len(results) == 2
    assert results == CERT_LIST
    assert keyword_search(
        CERT_LIST,
        certificate__contains='encryption'
    ) == []


def test_parse_exception():
    with pytest.raises(ParseException) as e_info:
        raise ParseException('This is a parse exception')
    assert 'This is a parse exception' == str(e_info.value)


def test_skip_exception():
    with pytest.raises(SkipException) as e_info:
        raise SkipException('This is a skip exception')
    assert 'This is a skip exception' == str(e_info.value)



"""
OSADispatcherLog - file ``/var/log/rhn/osa-dispatcher.log``
===========================================================
"""
from insights.core.plugins import parser
from insights.core import LogFileOutput

from datetime import datetime
import re
from insights.specs import Specs


@parser(Specs.osa_dispatcher_log)
class OSADispatcherLog(LogFileOutput):
    """
    Reads the OSA dispatcher log.  Based on the ``LogFileOutput`` class.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`


    Works a bit like the XMLRPC log but the IP address always seems to be
    ``0.0.0.0`` and the module is always 'osad' - it's more like what
    produced the log.

    Sample log data::

        2015/12/23 04:40:58 -04:00 28307 0.0.0.0: osad/jabber_lib.__init__
        2015/12/23 04:40:58 -04:00 28307 0.0.0.0: osad/jabber_lib.setup_connection('Connected to jabber server', u'example.com')
        2015/12/23 04:40:58 -04:00 28307 0.0.0.0: osad/osa_dispatcher.fix_connection('Upstream notification server started on port', 1290)
        2015/12/23 04:40:58 -04:00 28307 0.0.0.0: osad/jabber_lib.process_forever
        2015/12/27 22:48:50 -04:00 28307 0.0.0.0: osad/jabber_lib.main('ERROR', 'Error caught:')
        2015/12/27 22:48:50 -04:00 28307 0.0.0.0: osad/jabber_lib.main('ERROR', 'Traceback (most recent call last)')

    Example:
        >>> osa = shared[OSADispatcherLog]
        >>> osa.get('__init__')
        [{'raw_message': '2015/12/23 04:40:58 -04:00 28307 0.0.0.0: osad/jabber_lib.__init__',
         'timestamp': '2015/12/23 04:40:58 -04:00',
         'datetime': datetime.datetime(2015, 12, 23, 4, 40, 58),
         'pid': '28307', 'client_ip': '0.0.0.0', 'module': 'osad',
         'function': 'jabber_lib.__init__', 'info': None}
        ]
        >>> osa.last()
        {'raw_message': "2015/12/27 22:48:50 -04:00 28307 0.0.0.0: osad/jabber_lib.main('ERROR', 'Traceback (most recent call last)')",
         'timestamp': '2015/12/27 22:48:50 -04:00',
         'datetime': datetime.datetime(2015, 12, 27, 22, 48, 50), 'pid': '28307',
         'client_ip': '0.0.0.0', 'module': 'osad', 'function': 'jabber_lib.main',
         'info': "'ERROR', 'Traceback (most recent call last)'"}
        >>> from datetime import datetime
        >>> len(list(osa.get_after(datetime(2015, 12, 27, 22, 48, 0))))
        2
    """
    time_format = '%Y/%m/%d %H:%M:%S'

    _LINE_STR = r"^(?P<timestamp>\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} " + \
        r"[+-]?\d{2}:\d{2}) (?P<pid>\d+) (?P<client_ip>\S+): " + \
        r"(?P<module>\w+)/(?P<function>[\w_.-]+)(?:\((?P<info>.*)\))?$"
    _LINE_RE = re.compile(_LINE_STR)

    def _parse_line(self, line):
        """
        Parse a log line using the XMLRPC regular expression into a dict.
        All data will be in fields, and the raw log line is stored in
        'raw_message'.

        This also attempts to convert the timestamp given into a datetime
        object; if it can't convert it, then you don't get a 'datetime'
        key in the line's dict.
        """
        msg_info = dict()
        msg_info['raw_message'] = line

        match = self._LINE_RE.search(line)
        if match:
            msg_info.update(match.groupdict())
            try:
                stamp = match.group('timestamp')
                # Must remove : from timezone for strptime %z
                msg_info['datetime'] = datetime.strptime(
                    stamp[0:23] + stamp[24:26], self.time_format + ' %z')
            except:
                pass

        return msg_info

    def last(self):
        """
        Finds the last complete log line in the file.  It looks for a line with
        a client IP address and parses the line to a dictionary.

        Returns:
            (dict) the last complete log line parsed to a dictionary.
        """
        msg_info = dict()
        for l in reversed(self.lines):
            msg_info = self._parse_line(l)
            # assume parse is successful if we got an IP address
            if 'client_ip' in msg_info:
                return msg_info
        # Return the last one even if it didn't parse.
        return msg_info



"""
RcLocal - file ``/etc/rc.d/rc.local``
=====================================
"""

from .. import Parser, parser, get_active_lines
from insights.specs import Specs


@parser(Specs.rc_local)
class RcLocal(Parser):
    """
    Parse the `/etc/rc.d/rc.local` file.

    Sample input::

        #!/bin/sh
        #
        # This script will be executed *after* all the other init scripts.
        # You can put your own initialization stuff in here if you don't
        # want to do the full Sys V style init stuff.

        touch /var/lock/subsys/local
        echo never > /sys/kernel/mm/redhat_transparent_hugepage/enabled

    Attributes
    ----------
    data: list
        List of all lines from `rc.local` that are not comments or blank

    Examples
    --------
    >>> shared[RcLocal].data[0]
    'touch /var/lock/subsys/local'
    >>> shared[RcLocal].get('kernel')
    ['echo never > /sys/kernel/mm/redhat_transparent_hugepage/enabled']

    """
    def parse_content(self, content):
        self.data = [l for l in get_active_lines(content)]

    def get(self, value):
        """Returns the lines containing string value."""
        return [l for l in self.data if value in l]



"""
PodmanList - command ``/usr/bin/podman (images|ps)``
====================================================

Parse the output of command "podman_list_images" and "podman_list_containers",
which have very similar formats.

The header line is parsed and used as the names for the remaining columns.
All fields in both header and data are assumed to be separated by at least
three spaces. This allows single spaces in values and headers, so headers
such as 'IMAGE ID' are captured as is.

If the header line and at least one data line are not found, no data is
stored.

Each row is stored as a dictionary, keyed on the header fields. The data is
available in two formats:

* The old format is a list of row dictionaries.
* The new format stores each dictionary in a dictionary keyed on the value of
  a given field, given by the subclass.

"""
from insights import CommandParser, parser
from insights.parsers import SkipException, parse_fixed_table
from insights.specs import Specs


class PodmanList(CommandParser):
    """
    A general class for parsing tabular podman list information.  Parsing
    rules are:

    * The first line is the header line.
    * The other lines are data lines.
    * All fields line up vertically.
    * Fields are separated from each other by at least three spaces.
    * Some fields can contain nothing, and this is shown as spaces, so we
      need to catch that and turn it into None.

    Why not just use hard-coded fields and columns?  So that we can adapt to
    different output lists.

    Raises:
        NotImplementedError: If `key_field` or `attr_name` is not defined
        SkipException: If no data to parse
    """
    key_field = ''
    heading_ignore = []
    attr_name = ''
    substitutions = []

    def parse_content(self, content):
        if not (self.key_field and self.attr_name):
            raise NotImplementedError("'key_field' or 'attr_name' is not defined")

        self.rows = parse_fixed_table(content,
                                      heading_ignore=self.heading_ignore,
                                      header_substitute=self.substitutions)

        if not self.rows:
            raise SkipException('No data.')

        data = {}
        for row in self.rows:
            k = row.get(self.key_field)
            for sub in self.substitutions:
                row[sub[0]] = row.pop(sub[1])
            if k is not None and k != '<none>':
                data[k] = row
        setattr(self, self.attr_name, data)


@parser(Specs.podman_list_images)
class PodmanListImages(PodmanList):
    """
    Handle the list of podman images using the PodmanList parser class.

    Sample output of command ``podman images --all --no-trunc --digests``::

        REPOSITORY                           TAG                 DIGEST              IMAGE ID                                                           CREATED             SIZE
        rhel6_vsftpd                         latest              <none>              412b684338a1178f0e5ad68a5fd00df01a10a18495959398b2cf92c2033d3d02   37 minutes ago      459.5 MB
        rhel7_imagemagick                    latest              <none>              882ab98aae5394aebe91fe6d8a4297fa0387c3cfd421b2d892bddf218ac373b2   4 days ago          785.4 MB
        rhel6_nss-softokn                    latest              <none>              dd87dad2c7841a19263ae2dc96d32c501ee84a92f56aed75bb67f57efe4e48b5   5 days ago          449.7 MB

    Attributes:
        rows (list): List of row dictionaries.
        images (dict): Dictionary keyed on the value of the "REPOSITORY" fileld

    Examples:
        >>> images.rows[0]['REPOSITORY']
        'rhel6_vsftpd'
        >>> images.rows[1]['SIZE']
        '785.4 MB'
        >>> images.images['rhel6_vsftpd']['CREATED']
        '37 minutes ago'
    """
    key_field = 'REPOSITORY'
    heading_ignore = [key_field]
    attr_name = 'images'
    substitutions = [("IMAGE ID", "IMAGE_ID")]


@parser(Specs.podman_list_containers)
class PodmanListContainers(PodmanList):
    """
    Handle the list of podman containers using the PodmanList parser class.


    Sample output of command ``podman ps --all --no-trunc --size``::

        CONTAINER ID                                                       IMAGE                                                              COMMAND                                            CREATED             STATUS                        PORTS                  NAMES               SIZE
        03e2861336a76e29155836113ff6560cb70780c32f95062642993b2b3d0fc216   rhel7_httpd                                                        "/usr/sbin/httpd -DFOREGROUND"                     45 seconds ago      Up 37 seconds                 0.0.0.0:8080->80/tcp   angry_saha          796 B (virtual 669.2 MB)
        95516ea08b565e37e2a4bca3333af40a240c368131b77276da8dec629b7fe102   bd8638c869ea40a9269d87e9af6741574562af9ee013e03ac2745fb5f59e2478   "/bin/sh -c 'yum install -y vsftpd-2.2.2-6.el6'"   51 minutes ago      Exited (137) 50 minutes ago                          tender_rosalind     4.751 MB (virtual 200.4 MB)

    Attributes:
        rows (list): List of row dictionaries.
        containers(dict): Dictionary keyed on the value of the "NAMES" field

    Examples:
        >>> containers.rows[0]['NAMES']
        'angry_saha'
        >>> containers.rows[0]['STATUS']
        'Up 37 seconds'
        >>> containers.containers['tender_rosalind']['STATUS']
        'Exited (137) 18 hours ago'
    """
    key_field = 'NAMES'
    heading_ignore = ['CONTAINER']
    attr_name = 'containers'
    substitutions = [("CONTAINER ID", "CONTAINER_ID")]



"""
MaxUID - command ``/bin/awk -F':' '{ if($3 > max) max = $3 } END { print max }' /etc/passwd``
=============================================================================================

This module provides the MaxUID value gathered from the ``/etc/passwd`` file.
"""

from insights.core import Parser
from insights.core.plugins import parser
from insights.parsers import ParseException, SkipException
from insights.specs import Specs


@parser(Specs.max_uid)
class MaxUID(Parser):
    """
    Class for parsing the MaxUID value from the ``/etc/passwd`` file returned by the command::

        /bin/awk -F':' '{ if($3 > max) max = $3 } END { print max }' /etc/passwd

    Typical output of the ``/etc/passwd`` file is::

        root:x:0:0:root:/root:/bin/bash
        bin:x:1:1:bin:/bin:/sbin/nologin
        daemon:x:2:2:daemon:/sbin:/sbin/nologin
        adm:x:3:4:adm:/var/adm:/sbin/nologin
        lp:x:4:7:lp:/var/spool/lpd:/sbin/nologin
        sync:x:5:0:sync:/sbin:/bin/sync
        shutdown:x:6:0:shutdown:/sbin:/sbin/shutdown
        halt:x:7:0:halt:/sbin:/sbin/halt
        mail:x:8:12:mail:/var/spool/mail:/sbin/nologin
        nobody:x:65534:65534:Kernel Overflow User:/:/sbin/nologin

    Typical output of this parser is::

        65534

    Raises:
        SkipException: When content is empty or cannot be parsed.
        ParseException: When type cannot be recognized.

    Examples:
        >>> max_uid.value
        65534
    """

    def parse_content(self, content):
        if not content:
            raise SkipException("No content.")

        for line in content:
            try:
                self.value = int(line)
            except ValueError as e:
                raise ParseException("Failed to parse content with error: {}", format(str(e)))



"""
PackageProvidesHttpdAll - Combiner for packages which provide httpd
===================================================================

Combiner for collecting all the running httpd command and the corresponding RPM package name
which is parsed by the PackageProvidesHttpd parser.

"""

from insights.core.plugins import combiner
from insights.parsers.package_provides_httpd import PackageProvidesHttpd
from .. import LegacyItemAccess


@combiner(PackageProvidesHttpd)
class PackageProvidesHttpdAll(LegacyItemAccess):
    """
    This combiner will receive a list of parsers named PackageProvidesHttpd, one for each running instance of httpd
    and each parser instance will contain the command information and the RPM package information.
    It works as a ``dict`` with the httpd command information as the key and the
    corresponding RPM package information as the value.

    Examples:
        >>> sorted(packages.running_httpds)
        ['/opt/rh/httpd24/root/usr/sbin/httpd', '/usr/sbin/httpd']
        >>> packages.get_package("/usr/sbin/httpd")
        'httpd-2.4.6-88.el7.x86_64'
        >>> packages.get("/opt/rh/httpd24/root/usr/sbin/httpd")
        'httpd24-httpd-2.4.34-7.el7.x86_64'
        >>> packages["/usr/sbin/httpd"]
        'httpd-2.4.6-88.el7.x86_64'

    """

    def __init__(self, package_provides_httpd):
        self.data = {}
        for pkg in package_provides_httpd:
            self.data[pkg.command] = pkg.package
        super(PackageProvidesHttpdAll, self).__init__()

    @property
    def running_httpds(self):
        """
        Returns the list of httpd commands which are running on the system.
        """
        return list(self.data.keys())

    def get_package(self, httpd_command):
        """
        Returns the installed httpd package that provides the specified `httpd_command`.

        Parameters:
            httpd_command (str): The specified httpd command, e.g. found in ``ps`` command.

        Returns:
            (str): The package that provides the httpd command.
        """

        return self.data.get(httpd_command)

    @property
    def packages(self):
        """
        Returns the list of corresponding httpd RPM packages which are running on the system.
        """
        return list(self.data.values())



"""
Top level OpenShift 4 component
===============================
The :py:func:`conf` component recognizes insights-operator and must-gather
archives.
"""
import logging
import os
import yaml

from fnmatch import fnmatch
from insights.core.plugins import component, datasource
from insights.core.context import InsightsOperatorContext, MustGatherContext

from insights.core.archives import extract
from insights.parsr.query import from_dict, Result
from insights.util import content_type


log = logging.getLogger(__name__)
contexts = [InsightsOperatorContext, MustGatherContext]

try:
    # requires pyyaml installed after libyaml
    Loader = yaml.CSafeLoader
except:
    log.info("Couldn't find libyaml loader. Falling back to python loader.")
    Loader = yaml.SafeLoader


def _get_files(path):
    for root, dirs, names in os.walk(path):
        for name in names:
            yield os.path.join(root, name)


def _load(path):
    with open(path) as f:
        doc = yaml.load(f, Loader=Loader)
        return from_dict(doc)


def _process(path, excludes=None):
    excludes = excludes if excludes is not None else []
    for f in _get_files(path):
        if excludes and any(fnmatch(f, e) for e in excludes):
            continue
        try:
            yield _load(f)
        except Exception:
            log.debug("Failed to load %s; skipping.", f)


def analyze(paths, excludes=None):
    if not isinstance(paths, list):
        paths = [paths]

    results = []
    for path in paths:
        if content_type.from_file(path) == "text/plain":
            results.append(_load(path))
        elif os.path.isdir(path):
            results.extend(_process(path, excludes))
        else:
            with extract(path) as ex:
                results.extend(_process(ex.tmp_dir, excludes))

    return Result(children=results)


@datasource(contexts)
def conf_root(broker):
    for ctx in contexts:
        if ctx in broker:
            return broker[ctx].root


@component(conf_root)
def conf(root):
    """
    The ``conf`` component parses all configuration in an insights-operator or
    must-gather archive and returns an object that is part of the parsr common
    data model.  It can be navigated and queried in a standard way. See the
    `tutorial`_ for details.

    .. _tutorial: https://insights-core.readthedocs.io/en/latest/notebooks/Parsr%20Query%20Tutorial.html
    """
    return analyze(root, excludes=["*.log"])



from insights.parsers.tmpfilesd import TmpFilesD
from insights.combiners.tmpfilesd import AllTmpFiles
from insights.tests import context_wrap

SAP_CONF = """
# systemd tmpfiles exclude file for SAP
# SAP software stores some important files
# in /tmp which should not be deleted

# Exclude SAP socket and lock files
x /tmp/.sap*

# Exclude HANA lock file
x /tmp/.hdb*lock
""".strip()

DUPE_SAP_CONF = """
# systemd tmpfiles exclude file for SAP
# SAP software stores some important files
# in /tmp which should not be deleted

# Exclude SAP socket and lock files
r! /tmp/.sap*

# Exclude HANA lock file
x /tmp/.hdb*lock
""".strip()

DNF_CONF = """
# /usr/lib/tmpfiles.d/dnf.conf
r! /var/cache/dnf/*/*/download_lock.pid
r! /var/cache/dnf/*/*/metadata_lock.pid
r! /var/lib/dnf/rpmdb_lock.pid
e  /var/cache/dnf/ - - - 3
""".strip()

SAP_PATH = "/etc/tmpfiles.d/sap.conf"
DNF_PATH = "/etc/tmpfiles.d/dnf.conf"
DUPE_PATH = "/usr/lib/tmpfiles.d/sap.conf"


def test_all_tmpfiles_conf():
    data1 = TmpFilesD(context_wrap(SAP_CONF, SAP_PATH))
    data2 = TmpFilesD(context_wrap(DNF_CONF, DNF_PATH))
    all_data = AllTmpFiles([data1, data2])

    assert len(all_data.active_rules) == 2
    assert all_data.active_rules['/etc/tmpfiles.d/sap.conf'] == [{'type': 'x', 'path': '/tmp/.sap*', 'age': None,
                                                                  'gid': None, 'uid': None, 'argument': None, 'mode': None},
                                                                 {'type': 'x', 'path': '/tmp/.hdb*lock', 'age': None,
                                                                  'gid': None, 'uid': None, 'argument': None, 'mode': None}]
    assert all_data.active_rules['/etc/tmpfiles.d/dnf.conf'] == [{'type': 'r!', 'path': '/var/cache/dnf/*/*/download_lock.pid',
                                                                  'age': None, 'gid': None, 'uid': None, 'argument': None,
                                                                  'mode': None},
                                                                 {'type': 'r!', 'path': '/var/cache/dnf/*/*/metadata_lock.pid',
                                                                  'age': None, 'gid': None, 'uid': None, 'argument': None,
                                                                  'mode': None},
                                                                 {'type': 'r!', 'path': '/var/lib/dnf/rpmdb_lock.pid',
                                                                  'age': None, 'gid': None, 'uid': None, 'argument': None,
                                                                  'mode': None},
                                                                 {'type': 'e', 'path': '/var/cache/dnf/', 'age': '3',
                                                                  'gid': '-', 'uid': '-', 'argument': None, 'mode': '-'}]

    assert all_data.find_file('/tmp/.sap*') == {SAP_PATH: {'type': 'x', 'path': '/tmp/.sap*', 'age': None,
                                                           'mode': None, 'gid': None, 'uid': None,
                                                           'argument': None}}

    assert all_data.find_file('/tmp/.sap*')[SAP_PATH]['type'] == 'x'

    data3 = TmpFilesD(context_wrap(DUPE_SAP_CONF, DUPE_PATH))
    all_data = AllTmpFiles([data1, data3])
    # Ensure that only the last discovered of the rules are counted
    assert len(all_data.active_rules) == 2
    assert all_data.active_rules['/etc/tmpfiles.d/sap.conf'][0] == {'type': 'x', 'path': '/tmp/.sap*', 'age': None,
                                                                    'gid': None, 'uid': None, 'argument': None, 'mode': None}

    all_data = AllTmpFiles([data1, data2, data3])
    assert len(all_data.file_paths) == 3
    assert all_data.file_paths == ['/etc/tmpfiles.d/dnf.conf',
                                   '/etc/tmpfiles.d/sap.conf',
                                   '/usr/lib/tmpfiles.d/sap.conf']

    assert len(all_data.files) == 6
    assert all_data.files == set(['/var/cache/dnf/*/*/download_lock.pid',
                                  '/var/cache/dnf/*/*/metadata_lock.pid',
                                  '/var/lib/dnf/rpmdb_lock.pid',
                                  '/var/cache/dnf/',
                                  '/tmp/.sap*',
                                  '/tmp/.hdb*lock'])



import doctest

from insights.parsers import ironic_conf
from insights.parsers.ironic_conf import IronicConf
from insights.tests import context_wrap

ironic_content = """
[DEFAULT]
auth_strategy=keystone
default_resource_class=baremetal
enabled_hardware_types=idrac,ilo,ipmi,redfish
enabled_bios_interfaces=no-bios
enabled_boot_interfaces=ilo-pxe,pxe
enabled_console_interfaces=ipmitool-socat,ilo,no-console
enabled_deploy_interfaces=iscsi,direct,ansible
enabled_inspect_interfaces=inspector,no-inspect
default_inspect_interface=inspector
enabled_management_interfaces=fake,idrac,ilo,ipmitool,noop,redfish
enabled_network_interfaces=flat
default_network_interface=flat
enabled_power_interfaces=fake,idrac,ilo,ipmitool,redfish
enabled_raid_interfaces=idrac,no-raid
enabled_rescue_interfaces=no-rescue,agent
default_rescue_interface=agent
enabled_storage_interfaces=noop
enabled_vendor_interfaces=idrac,ipmitool,no-vendor
my_ip=192.168.24.1
debug=True
log_dir=/var/log/ironic
[agent]
deploy_logs_collect=always
deploy_logs_storage_backend=local
deploy_logs_local_path=/var/log/ironic/deploy/
[ansible]
[api]
[audit]
[cimc]
[cinder]
auth_type=password
password=Nzmsp57Yg94HXsv9x8Znc7tEI
project_domain_name=Default
project_name=service
user_domain_name=Default
username=ironic
[cisco_ucs]
[conductor]
force_power_state_during_sync=False
automated_clean=False
max_time_interval=120
[console]
[cors]
allowed_origin=*
expose_headers=Content-Type,Cache-Control,Content-Language,Expires,Last-Modified,Pragma
max_age=3600
allow_methods=GET,POST,PUT,DELETE,OPTIONS,PATCH
allow_headers=Content-Type,Cache-Control,Content-Language,Expires,Last-Modified,Pragma,X-Auth-Token
[database]
[deploy]
http_root=/var/lib/ironic/httpboot
erase_devices_priority=0
erase_devices_metadata_priority=10
"""


def test_ironic_conf():
    result = IronicConf(context_wrap(ironic_content))
    assert result.get("DEFAULT", "auth_strategy") == "keystone"
    assert result.get("DEFAULT", "default_resource_class") == "baremetal"
    assert result.get("DEFAULT", "default_rescue_interface") == "agent"
    assert result.get("agent", "deploy_logs_collect") == "always"
    assert result.get("conductor", "force_power_state_during_sync") == "False"


def test_ironic_conf_docs():
    failed, total = doctest.testmod(
        ironic_conf,
        globs={
            'ironic_conf': IronicConf(context_wrap(ironic_content)),
        }
    )
    assert failed == 0



import os
from functools import reduce


def path_join(path1, path2):
    # os.path.join drops path1 entirely if path2 starts with a '/' which makes sense
    #    in some circomstances, but almost never in this program
    # so we use path_join instead
    return os.path.join(path1, path2 if not path2.startswith('/') else path2.lstrip('/'))


def sh_join(split_command):
    # is the reverse of shlex.split
    #   such that for some shell command string X:
    #     shlex.split(sh_join(shlex.split(X))) == shlex.split(X)
    # The point is that you can parse a command string with shlex.split,
    #   manipulate the individual arguments in the resulting array
    #   and then join the individual arguments back together in a way that a future
    #   shlex.split will properly parse the shell command string.
    # Most of the real work happens in sh_quote

    return reduce(lambda accum, arg: accum + ' ' + sh_quote(arg), split_command)


def sh_quote(arg):
    """
    will do shell style quoting for a single shell argument
      such that it can be recombined back into a single string (rather than an array of strings)
      such that shlex.split would resplit the string back into the same array.
    The characters that need quoting within an argument are: double-quote, single-quote, backslash,
      and all the whitespace characters (blank, newline, tab)
    If an arg has none of these, this function returns it as is.
    The simplest implementation of this function would simply put a backslash in front of
      any character that needs special quoting, but that would look really ugly in some cases,
      and since these modified commands end up in the uploader.json file, we try for something
      prettier.
    For the unix shell,
       single-quotes around, quotes everything, but can not contain a quote (this is different
         than Python string literals).
       double-quotes around, quotes single-quotes and whitespace, but double-quotes and backslashes
         must be escaped with a backslash.
       outside of quotes, backslash always escapes the following character

    The way we do this is to count the characters that need quoteing, and ....
    This is simpler to understand, but does at least 6 passes over the string, and can do many
      more.  This generally isn't a problem because most arguments are very short.
    A faster way to implement this would be to do one pass over the string, start building an
      output string as soon as it becomes needed, and stoping once an output string can't be used.
      If more than possible output string is left at the end, choose the best.
      For example, until you see a special character, only the unmodified output string is needed
      As soon as a space is encountered, create the single-quote output, double-quote output,
        and backslash output strings, and drop the unmodified output string.
      As soon as a single-quote is found drop the single-quote output string.
      Etc.

    But for now, do it the slower, simple way.
    """
    double_count = arg.count('"')
    single_count = arg.count("'")
    space_count = arg.count(' ') + arg.count('\t') + arg.count('\n')
    backslash_count = arg.count('\\')

    # To try to make the output as pretty as we can we use the following:
    # if no characters need quoteing, return arg as is
    if double_count + single_count + space_count + backslash_count == 0:
        return arg

    # elif arg contains no single-quotes, return single-quotes around arg
    elif single_count == 0:
        return sh_single_quote(arg)

    # elif we need to quote only 1 or 2 character, use backslashes in front of
    elif double_count + single_count + space_count + backslash_count < 3:
        return sh_backslash_quote(arg)

    # else use double-quotes around
    else:
        return sh_double_quote(arg)


def sh_single_quote(arg):
    # put single-quotes around
    #    this will fail to quote correctly if arg contains any single-quotes
    return "'" + arg + "'"


def sh_double_quote(arg):
    # put double-quotes around, and escape any double-quotes and backslashes within arg
    #   this will always quote correctly
    return '"' + arg.replace('\\', '\\\\').replace('"', '\\"') + '"'


def sh_backslash_quote(arg):
    # quote each special character individually by putting a backslash in front of it.
    #   this will always quote correctly, but is ugly unless only a few characters need quoteing
    # Note that when doing the replaceing, replace the backslashes first or this will fail!
    return arg.replace('\\', '\\\\').replace('"', '\\"').replace("'", "\\'").replace(' ', '\\ ').replace('\t', '\\\t').replace('\n', '\\\n')



"""
LsVarRun - command ``ls -lnL /var/run``
=======================================

The ``ls -lnL /var/run`` command provides information for the listing of the
``/var/run`` directory.

Sample input is shown in the Examples. See ``FileListing`` class for
additional information.

Sample directory list::

    total 20
    drwx--x---.  2   0 984   40 May 15 09:29 openvpn
    drwxr-xr-x.  2   0   0   40 May 15 09:30 plymouth
    drwxr-xr-x.  2   0   0   40 May 15 09:29 ppp
    drwxr-xr-x.  2  75  75   40 May 15 09:29 radvd
    -rw-r--r--.  1   0   0    5 May 15 09:30 rhnsd.pid
    drwxr-xr-x.  2   0   0   60 May 30 09:31 rhsm
    drwx------.  2  32  32   40 May 15 09:29 rpcbind
    -r--r--r--.  1   0   0    0 May 17 16:26 rpcbind.lock

Examples:

    >>> "rhnsd.pid" in ls_var_run
    False
    >>> "/var/run" in ls_var_run
    True
    >>> ls_var_run.dir_entry('/var/run', 'openvpn')['type']
    'd'
"""


from insights.specs import Specs

from .. import FileListing
from .. import parser


@parser(Specs.ls_var_run)
class LsVarRun(FileListing):
    """Parses output of ``ls -lnL /var/run`` command."""
    pass
